import tsplib95 as tsplib

from tsplib95 import transformers, distances


def exact(start, end, scale=1000):
    start_scaled = [e * scale for e in start]
    end_scaled = [e * scale for e in end]
    return distances.euclidean(start_scaled, end_scaled)


distances.TYPES.update({
    'EXACT_2D': exact,
    'EXACT_3D': exact
})


class NodeListField(tsplib.fields.TransformerField):
    default = list

    @classmethod
    def build_transformer(cls):
        depot = transformers.FuncT(func=int)
        return transformers.ListT(value=depot, terminal='-1', sep='\n')


# format according to http://webhotel4.ruc.dk/~keld/research/LKH-3/LKH-3_REPORT.pdf
class LKHProblem(tsplib.models.StandardProblem):
    # extra spec fields
    salesmen = tsplib.fields.IntegerField('SALESMEN')
    vehicles = tsplib.fields.IntegerField('VEHICLES')
    distance = tsplib.fields.NumberField('DISTANCE')
    risk_threshold = tsplib.fields.IntegerField('RISK_THRESHOLD')
    scale = tsplib.fields.IntegerField('SCALE')

    # extra data fields
    backhaul = NodeListField('BACKHAUL_SECTION')
    pickup_and_delivery = tsplib.fields.MatrixField('PICKUP_AND_DELIVERY_SECTION')
    service_time = tsplib.fields.MatrixField('SERVICE_TIME_SECTION')
    time_window = tsplib.fields.MatrixField('TIME_WINDOW_SECTION')

    depots = NodeListField('DEPOT_SECTION')  # fix for https://github.com/rhgrant10/tsplib95/pull/16

    # need to override `render` because spec fields must precede data fields according to TSPLIB format
    # adapted from https://github.com/rhgrant10/tsplib95/blob/master/tsplib95/models.py#L217
    def render(self):
        spec_part = [
            'NAME',
            'TYPE',
            'COMMENT',
            'DIMENSION',
            'CAPACITY',
            'EDGE_WEIGHT_TYPE',
            'EDGE_WEIGHT_FORMAT',
            'EDGE_DATA_FORMAT',
            'NODE_COORD_TYPE',
            'DISPLAY_DATA_TYPE',
            'SALESMEN',
            'VEHICLES',
            'DISTANCE',
            'RISK_THRESHOLD',
            'SCALE'
        ]

        data_part = [
            'NODE_COORD_SECTION',
            'DEPOT_SECTION',
            'DEMAND_SECTION',
            'EDGE_DATA_SECTION',
            'FIXED_EDGES_SECTION',
            'DISPLAY_DATA_SECTION',
            'TOUR_SECTION',
            'EDGE_WEIGHT_SECTION',
            'BACKHAUL_SECTION',
            'PICKUP_AND_DELIVERY_SECTION',
            'SERVICE_TIME_SECTION',
            'TIME_WINDOW_SECTION'
        ]

        # render each value by keyword
        rendered = self.as_name_dict()
        for name in list(rendered):
            value = rendered.pop(name)
            field = self.__class__.fields_by_name[name]
            if name in self.__dict__ or value != field.get_default_value():
                rendered[field.keyword] = field.render(value)

        # build keyword-value pairs with the separator
        # add fields from spec part, then data part
        kvpairs = []
        for keyword in spec_part + data_part:
            if keyword in rendered:
                value = rendered[keyword]
                sep = ':\n' if '\n' in value else ': '
                kvpairs.append(f'{keyword}{sep}{value}')
        kvpairs.append('EOF')

        # join and return the result
        return '\n'.join(kvpairs)



import os
import shutil
import subprocess
import tempfile
import warnings

from .problems import LKHProblem


def solve(solver='LKH', problem=None, **params):
    assert shutil.which(solver) is not None, f'{solver} not found.'
    assert ('problem_file' in params) ^ (problem is not None), 'Specify a problem object *or* a path.'

    if 'problem_file' in params:
        # annoying, but necessary to get the original problem dimension
        problem = LKHProblem.load(params['problem_file'])

    if not isinstance(problem, LKHProblem):
        warnings.warn('Subclassing LKHProblem is recommended. Proceed at your own risk!')

    if len(problem.depots) > 1:
        warnings.warn('LKH-3 cannot solve multi-depot problems.')

    prob_file = tempfile.NamedTemporaryFile(mode='w', delete=False)
    problem.write(prob_file)
    prob_file.write('\n')
    prob_file.close()
    params['problem_file'] = prob_file.name

    if 'tour_file' not in params:
        tour_file = tempfile.NamedTemporaryFile(mode='w', delete=False)
        params['tour_file'] = tour_file.name
        tour_file.close()

    par_file = tempfile.NamedTemporaryFile(mode='w+', delete=False)
    par_file.write('SPECIAL\n')
    for k, v in params.items():
        par_file.write(f'{k.upper()} = {v}\n')
    par_file.close()

    try:
        # stdin=DEVNULL for preventing a "Press any key" pause at the end of execution
        subprocess.check_output([solver, par_file.name], stderr=subprocess.STDOUT, stdin=subprocess.DEVNULL)
    except subprocess.CalledProcessError as e:
        raise Exception(e.output.decode())

    # the tour file produced by LKH-3 includes dummy nodes to indicate depots
    # for example, if a problem has DIMENSION=32 (1 depot node + 31 task nodes),
    # the tour file will have a SINGLE tour with DIMENSION=36 (5 depot nodes + 31 task nodes)
    solution = LKHProblem.load(params['tour_file'])
    tour = solution.tours[0]
    # convert this tour to multiple routes
    routes = []
    route = []
    for node in tour:
        if node in problem.depots or node > problem.dimension:
            if len(route) > 0:
                routes.append(route)
            route = []
        else:
            route.append(node)
    routes.append(route)

    os.remove(par_file.name)
    if 'prob_file' in locals():
        os.remove(prob_file.name)
    if 'tour_file' in locals():
        os.remove(tour_file.name)

    return routes



from setuptools import setup, find_packages
from os import path

readme_path = path.join(path.abspath(path.dirname(__file__)), 'README.md')
with open(readme_path) as readme:
    long_description = readme.read()

setup(name='lkh',
    description='Super simple Python wrapper for LKH-3',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Ben Hudson',
    author_email='benhudson@fastmail.com',
    version='1.1.1',
    license='DBAD',
    packages=find_packages(),
    url='https://github.com/ben-hudson/pylkh',
    download_url='https://github.com/ben-hudson/pylkh',
    keywords=['TSP', 'CVRP', 'VRP', 'LKH', 'LKH-3'],
    python_requires='>=3.3',
    install_requires=[
        'tsplib95'
    ],
)
